import ray
import time

@ray.remote
class PushSendReceive(object):
    """
    send and receive initiated by the sender.
    message is sent as an argument to the receive function.

    The interface is a little wonky.
      * set a message
      * then send and receive
    This is designed to mimic an MPI session in which the message content is in
    a buffer and then send and receive send the buffer.
    The receiver returns the received message.
    Printing the send output verifies the message:
        * send -> recv (message as argument)
        * recv returns message to send
        * send prints argument to verify receipt
    """
    msg = ""

    def __init__(self):
        None

    def set_message(self, msg):
        """set the message to be sent"""
        self.msg = msg

    def push_send(self, ooid):
        """synchronous send: only return after the remote has received the message."""
        # sleep to make sure both actors are started
        time.sleep(1)
        ### TODO invoke the remote recv function.

    def recv(self, msg):
        """Receive a message as function argument.
            Return the messge to verify receipt.
        """
        ### TODO

# Start the ray framework
ray.init(num_cpus=4, ignore_reinit_error=True)

# Create two actors to do paired messaging.
sr0 = PushSendReceive.remote()
sr1 = PushSendReceive.remote()

# set messages
roid0 = sr0.set_message.remote(f"push_send test from {sr0}")
roid1 = sr1.set_message.remote(f"push_send test from {sr1}")
ray.get(roid0)
ray.get(roid1)

# synchronized sends should succeed
print("Synchronized. One after another.")
roid0 = sr0.push_send.remote(sr1)
# Verify that the message was received.
print(ray.get(roid0))

roid1 = sr1.push_send.remote(sr0)
# Verify that the message was received.
print(ray.get(roid1))

# concurrent sends should result in a deadlock
print("Concurrent")
roid0 = sr0.push_send.remote(sr1)
roid1 = sr1.push_send.remote(sr0)

print(ray.get(roid0))
print(ray.get(roid1))

print("Didn't get here. Deadlock.")

ray.shutdown()
